NUCLEOCOUNTER_SOFTWARE_NAME = "NucleoView"
NUCLEOCOUNTER_DEVICE_TYPE = "dark field imager (cell counter)"
NUCLEOCOUNTER_DETECTION_TYPE = "dark field"

DEFAULT_ANALYST = "NC-200"
DEFAULT_MODEL_NUMBER = "NucleoCounter NC-200"

MEASUREMENT_AGG_DOCUMENT_CUSTOM_FIELDS = {
    "Protocol template purpose",
    "Protocol template filename",
    "Protocol template title",
    "Protocol adaptation purpose",
    "Protocol adaptation title",
    "Protocol adaptation filename",
    "Login ID",
    "Event log format",
}
